# 工具函数
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import minmax_scale, scale, maxabs_scale, robust_scale
from loguru import logger


def is_event_feature(df, col, **kwargs):
    """事件类因子的判断函数

    事件因子的特征：多头事件发生时，因子值为1；空头事件发生时，因子值为-1；其他情况，因子值为0。

    :param df: DataFrame
    :param col: str, 因子字段名称
    """
    unique_values = df[col].unique()
    return all([x in [0, 1, -1] for x in unique_values])


def rolling_corr(df, col1, col2, window=300, min_periods=100, **kwargs):
    """滚动计算两个序列的相关系数

    :param df: pd.DataFrame
    :param col1: str
    :param col2: str
    :param window: int, default 300, 滚动窗口大小, None表示扩展窗口
    :param min_periods: int, default 100, 最小观测数量, 用于计算相关系数的最小观测数量
    """
    if kwargs.get("copy", False):
        df = df.copy()

    assert isinstance(df, pd.DataFrame), "df must be pd.DataFrame"
    assert col1 in df.columns, f"{col1} not in df.columns"
    assert col2 in df.columns, f"{col2} not in df.columns"

    new_col = kwargs.get("new_col", f"{col1}_corr_{col2}")
    assert min_periods < window, "min_periods must be less than window"
    df[new_col] = df[col1].rolling(window=window, min_periods=min_periods).corr(df[col2]).fillna(0)
    return df


def rolling_rank(df: pd.DataFrame, col, window=300, min_periods=100, new_col=None, **kwargs):
    """计算序列的滚动排名

    :param df: pd.DataFrame, 待计算的数据
    :param col: str, 待计算的列
    :param window: int, 滚动窗口大小, 默认为300
    :param min_periods: int, 最小计算周期, 默认为100
    :param new_col: str, 新列名，默认为 None, 表示使用 f'{col}_rank' 作为新列名
    """
    if kwargs.get("copy", False):
        df = df.copy()

    min_periods = kwargs.get("min_periods", 2)
    new_col = new_col if new_col else f"{col}_rank"
    df[new_col] = df[col].rolling(window=window, min_periods=min_periods).rank(pct=True)
    df[new_col] = df[new_col].fillna(0)
    return df


def rolling_norm(df: pd.DataFrame, col, window=300, min_periods=100, new_col=None, **kwargs):
    """计算序列的滚动归一化值

    :param df: pd.DataFrame, 待计算的数据
    :param col: str, 待计算的列
    :param window: int, 滚动窗口大小, 默认为300
    :param min_periods: int, 最小计算周期, 默认为100
    :param new_col: str, 新列名，默认为 None, 表示使用 f'{col}_norm' 作为新列名
    """
    if kwargs.get("copy", False):
        df = df.copy()

    min_periods = kwargs.get("min_periods", 2)
    new_col = new_col if new_col else f"{col}_norm"
    df[new_col] = (
        df[col].rolling(window=window, min_periods=min_periods).apply(lambda x: (x[-1] - x.mean()) / x.std(), raw=True)
    )
    df[new_col] = df[new_col].fillna(0)
    return df


def rolling_qcut(df: pd.DataFrame, col, window=300, min_periods=100, new_col=None, **kwargs):
    """计算序列的滚动分位数

    :param df: pd.DataFrame, 待计算的数据
    :param col: str, 待计算的列
    :param window: int, 滚动窗口大小, 默认为300
    :param min_periods: int, 最小计算周期, 默认为100
    :param new_col: str, 新列名，默认为 None, 表示使用 f'{col}_qcut' 作为新列名
    """
    if kwargs.get("copy", False):
        df = df.copy()

    q = kwargs.get("q", 10)
    min_periods = kwargs.get("min_periods", q)
    new_col = new_col if new_col else f"{col}_qcut"

    def __qcut_func(x):
        return pd.qcut(x, q=q, labels=False, duplicates="drop")[-1]

    df[new_col] = df[col].rolling(window=window, min_periods=min_periods).apply(__qcut_func, raw=True)
    df[new_col] = df[new_col].fillna(-1)
    return df


def rolling_compare(df, col1, col2, window=300, min_periods=100, new_col=None, **kwargs):
    """计算序列的滚动归一化值

    :param df: pd.DataFrame
        待计算的数据
    :param col1: str
        第一个列名
    :param col2: str
        第二个列名
    :param window: int
        滚动窗口大小, 默认为300
    :param new_col: str
        新列名，默认为 None, 表示使用 f'{col}_norm' 作为新列名
    :param kwargs:
        min_periods: int
            最小计算周期
    """
    window = kwargs.get("window", 300)
    min_periods = kwargs.get("min_periods", 2)
    new_col = new_col if new_col else f"compare_{col1}_{col2}"
    method = kwargs.get("method", "sub")
    assert method in [
        "sub",
        "divide",
        "lr_intercept",
        "lr_coef",
    ], "method 必须为 sub, divide, lr_intercept, lr_coef 中的一种"

    for i in range(len(df)):
        dfi = df.loc[i - window + 1 : i, [col1, col2]]
        dfi = dfi.copy()
        if i < min_periods:
            df.loc[i, new_col] = 0
            continue

        if method == "sub":
            df.loc[i, new_col] = dfi[col1].sub(dfi[col2]).mean()

        elif method == "divide":
            df.loc[i, new_col] = dfi[col1].divide(dfi[col2]).mean()

        elif method == "lr_intercept":
            x = dfi[col2].values.reshape(-1, 1)
            y = dfi[col1].values.reshape(-1, 1)
            reg = LinearRegression().fit(x, y)
            df.loc[i, new_col] = reg.intercept_[0]

        elif method == "lr_coef":
            x = dfi[col2].values.reshape(-1, 1)
            y = dfi[col1].values.reshape(-1, 1)
            reg = LinearRegression().fit(x, y)
            df.loc[i, new_col] = reg.coef_[0][0]

        else:
            raise ValueError(f"method {method} not support")


def rolling_scale(df: pd.DataFrame, col: str, window=300, min_periods=100, new_col=None, **kwargs):
    """对序列进行滚动归一化

    :param df: pd.DataFrame, 待计算的数据
    :param col: str, 待计算的列
    :param window: int, 滚动窗口大小, 默认为300
    :param min_periods: int, 最小计算周期, 默认为100
    :param new_col: str, 新列名，默认为 None, 表示使用 f'{col}_scale' 作为新列名
    """
    if kwargs.get("copy", False):
        df = df.copy()

    df = df.sort_values("dt", ascending=True).reset_index(drop=True)
    new_col = new_col if new_col else f"{col}_scale"

    method = kwargs.get("method", "scale")
    method_map = {
        "scale": scale,
        "minmax_scale": minmax_scale,
        "maxabs_scale": maxabs_scale,
        "robust_scale": robust_scale,
    }
    assert method in method_map, f"method must be one of {list(method_map.keys())}"
    scale_method = method_map[method]

    if method == "minmax_scale":
        df[new_col] = (
            df[col]
            .rolling(window=window, min_periods=min_periods)
            .apply(lambda x: minmax_scale(x, feature_range=(-1, 1))[-1])
        )
    else:
        df[new_col] = df[col].rolling(window=window, min_periods=min_periods).apply(lambda x: scale_method(x)[-1])

    df[new_col] = df[new_col].fillna(0)
    return df


def rolling_tanh(df: pd.DataFrame, col: str, window=300, min_periods=100, new_col=None, **kwargs):
    """对序列进行滚动 tanh 变换

    双曲正切函数：https://baike.baidu.com/item/%E5%8F%8C%E6%9B%B2%E6%AD%A3%E5%88%87%E5%87%BD%E6%95%B0/15469414

    :param df: pd.DataFrame, 待计算的数据
    :param col: str, 待计算的列
    :param window: int, 滚动窗口大小, 默认为300
    :param min_periods: int, 最小计算周期, 默认为100
    :param new_col: str, 新列名，默认为 None, 表示使用 f'{col}_scale' 作为新列名
    """
    if kwargs.get("copy", False):
        df = df.copy()
    new_col = new_col if new_col else f"{col}_tanh"
    df = df.sort_values("dt", ascending=True).reset_index(drop=True)
    df[new_col] = df[col].rolling(window=window, min_periods=min_periods).apply(lambda x: np.tanh(scale(x))[-1])  # type: ignore
    df[new_col] = df[new_col].fillna(0)
    return df


def rolling_slope(df: pd.DataFrame, col: str, window=300, min_periods=100, new_col=None, **kwargs):
    """计算序列的滚动斜率

    大于0表示序列的斜率向上，小于0表示序列的斜率向下，绝对值越大表示斜率越陡峭

    :param df: pd.DataFrame, 待计算的数据
    :param col: str, 待计算的列
    :param window: int, 滚动窗口大小, 默认为300
    :param min_periods: int, 最小计算周期, 默认为100
    :param new_col: str, 新列名，默认为 None, 表示使用 f'{col}_slope' 作为新列名
    :param kwargs:

        - min_periods: int, 最小计算周期
        - method: str, 计算方法

            - linear: 使用线性回归计算斜率
            - std/mean: 使用序列的 std/mean 计算斜率
            - snr: 使用序列的 snr 计算斜率
    """
    method = kwargs.get("method", "linear")
    new_col = new_col if new_col else f"{col}_slope_{method}"

    if method == "linear":
        # 使用线性回归计算斜率
        def __lr_slope(x):
            return LinearRegression().fit(list(range(len(x))), x).coef_[0]

        df[new_col] = df[col].rolling(window=window, min_periods=min_periods).apply(__lr_slope, raw=True)

    elif method == "std/mean":
        # 用 window 内 std 的变化率除以 mean 的变化率，来衡量序列的斜率
        # 如果 std/mean > 0, 则表示序列的斜率在变大，反之则表示序列的斜率在变小
        df["temp_std"] = df[col].rolling(window=window, min_periods=min_periods).std().pct_change(window)
        df["temp_mean"] = df[col].rolling(window=window, min_periods=min_periods).mean().pct_change(window)
        df[new_col] = np.where(df["temp_mean"] != 0, df["temp_std"] / df["temp_mean"], 0)
        # 加入变化率的正负号
        df[new_col] = df[new_col] * np.sign(df[col].pct_change(window))
        df.drop(["temp_std", "temp_mean"], axis=1, inplace=True)

    elif method == "snr":
        # 用 window 内的信噪比变化率来衡量序列的斜率
        df[new_col] = df[col].diff(window) / df[col].diff().abs().rolling(window=window, min_periods=min_periods).sum()

    else:
        raise ValueError(f"Unknown method: {method}")

    df[new_col] = df[new_col].fillna(0)
    return df


def normalize_corr(df: pd.DataFrame, fcol, ycol=None, **kwargs):
    """标准化因子与收益相关性为正数

    方法说明：对因子进行滚动相关系数计算，因子乘以滚动相关系数的符号

    **注意：**

    1. simple 模式下，计算过程有一定的未来信息泄露，在回测中使用时需要注意
    2. rolling 模式下，计算过程依赖 window 参数，有可能调整后相关性为负数

    :param df: pd.DataFrame, 必须包含 dt、symbol、price 列，以及因子列
    :param fcol: str 因子列名
    :param kwargs: dict

        - window: int, 滚动窗口大小
        - min_periods: int, 最小计算周期
        - mode: str, 计算方法, rolling 表示使用滚动调整相关系数，simple 表示使用镜像反转相关系数
        - copy: bool, 是否复制 df

    :return: pd.DataFrame
    """
    window = kwargs.get("window", 1000)
    min_periods = kwargs.get("min_periods", 5)
    mode = kwargs.get("mode", "rolling")
    if kwargs.get("copy", False):
        df = df.copy()

    df = df.sort_values(["symbol", "dt"], ascending=True).reset_index(drop=True)
    for symbol, dfg in df.groupby("symbol"):
        dfg["ycol"] = dfg["price"].pct_change().shift(-1)

        if mode.lower() == "rolling":
            dfg["corr_sign"] = np.sign(dfg[fcol].rolling(window=window, min_periods=min_periods).corr(dfg["ycol"]))
            dfg[fcol] = (dfg["corr_sign"].shift(3) * dfg[fcol]).fillna(0)

        elif mode.lower() == "simple":
            corr_sign = np.sign(dfg[fcol].corr(dfg["ycol"]))
            dfg[fcol] = corr_sign * dfg[fcol]

        else:
            raise ValueError(f"Unknown mode: {mode}")

        df.loc[df["symbol"] == symbol, fcol] = dfg[fcol]
    return df


def feature_adjust_V230101(df: pd.DataFrame, fcol, **kwargs):
    """特征调整函数：对特征进行调整，使其符合持仓权重的定义

    方法说明：对因子进行滚动相关系数计算，然后对因子值用 maxabs_scale 进行归一化，最后乘以滚动相关系数的符号

    :param df: pd.DataFrame, 必须包含 dt、symbol、price 列，以及因子列
    :param fcol: str 因子列名
    :param kwargs: dict
    """
    window = kwargs.get("window", 1000)
    min_periods = kwargs.get("min_periods", 200)

    df = df.copy().sort_values("dt", ascending=True).reset_index(drop=True)
    df["n1b"] = df["price"].shift(-1) / df["price"] - 1
    df["corr"] = df[fcol].rolling(window=window, min_periods=min_periods).corr(df["n1b"])
    df["corr"] = df["corr"].shift(5).fillna(0)

    df = rolling_scale(
        df, col=fcol, window=window, min_periods=min_periods, new_col="weight", method="maxabs_scale", copy=True
    )
    df["weight"] = df["weight"] * np.sign(df["corr"])

    df.drop(["n1b", "corr"], axis=1, inplace=True)
    return df


def feature_adjust_V240323(df: pd.DataFrame, fcol, **kwargs):
    """特征调整函数：对特征进行调整，使其符合持仓权重的定义

    方法说明：对因子进行滚动相关系数计算，然后对因子值用 scale + tanh 进行归一化，最后乘以滚动相关系数的符号

    :param df: pd.DataFrame, 必须包含 dt、symbol、price 列，以及因子列
    :param fcol: str 因子列名
    :param kwargs: dict
    """
    window = kwargs.get("window", 1000)
    min_periods = kwargs.get("min_periods", 200)

    df = df.copy().sort_values("dt", ascending=True).reset_index(drop=True)
    df["n1b"] = df["price"].shift(-1) / df["price"] - 1
    df["corr"] = df[fcol].rolling(window=window, min_periods=min_periods).corr(df["n1b"])
    df["corr"] = df["corr"].shift(5).fillna(0)

    df = rolling_tanh(df, col=fcol, window=window, min_periods=min_periods, new_col="weight")
    df["weight"] = df["weight"] * np.sign(df["corr"])

    df.drop(["n1b", "corr"], axis=1, inplace=True)
    return df


def feature_adjust(df: pd.DataFrame, fcol, method, **kwargs):
    """特征调整函数：对特征进行调整，使其符合持仓权重的定义

    :param df: pd.DataFrame, 待调整的数据
    :param fcol: str, 因子列名
    :param method: str, 调整方法

        - KEEP: 直接使用原始因子值作为权重
        - V230101: 对因子进行滚动相关系数计算，然后对因子值用 maxabs_scale 进行归一化，最后乘以滚动相关系数的符号
        - V240323: 对因子进行滚动相关系数计算，然后对因子值用 scale + tanh 进行归一化，最后乘以滚动相关系数的符号

    :param kwargs: dict

        - window: int, 滚动窗口大小
        - min_periods: int, 最小计算周期

    :return: pd.DataFrame, 新增 weight 列
    """
    if method == "KEEP":
        df["weight"] = df[fcol]
        return df

    if method == "V230101":
        return feature_adjust_V230101(df, fcol, **kwargs)
    elif method == "V240323":
        return feature_adjust_V240323(df, fcol, **kwargs)
    else:
        raise ValueError(f"Unknown method: {method}")


def feature_to_weight(df, factor, positive, **kwargs):
    """时序因子转换为持仓权重

    处理流程：

    1. 缩尾处理：去除极端值
    2. scale 缩放，均值为0
    3. maxabs_scale 缩放至 [-1, 1]

    :param df: pd.DataFrame, 包含因子列的数据
    :param factor: str, 因子列名
    :param positive: bool, 因子是否为正向因子
    :param kwargs:

        - window: int, 计算窗口长度，默认为1000
        - min_periods: int, 最小计算窗口长度，默认为100
        - q_threshold: float, 缩尾阈值，默认为0.05

    """
    window = kwargs.get("window", 1000)
    min_periods = kwargs.get("min_periods", 100)
    q_threshold = kwargs.get("q_threshold", 0.05)
    assert df["symbol"].nunique() == 1, "必须按品种计算权重"

    # 缩尾处理
    df["upper"] = df[factor].rolling(window, min_periods).quantile(1 - q_threshold)
    df["lower"] = df[factor].rolling(window, min_periods).quantile(q_threshold)
    df[factor] = df[factor].clip(lower=df["lower"], upper=df["upper"])

    # scale 缩放，均值为0
    df["norm"] = df[factor].rolling(window, min_periods).apply(lambda x: scale(x)[-1])

    # maxabs_scale 缩放至 [-1, 1]
    df["weight"] = df["norm"].rolling(window, min_periods).apply(lambda x: maxabs_scale(x)[-1])
    df["weight"] = df["weight"].fillna(0)
    if not positive:
        df["weight"] = -df["weight"]

    df.drop(["upper", "lower", "norm"], axis=1, inplace=True)
    return df


def feature_returns(df, factor, target="n1b", **kwargs):
    """计算因子特征截面收益率

    :param df: pd.DataFrame, 必须包含 dt、symbol、factor, target 列
    :param factor: str, 因子列名
    :param target: str, 预测目标收益率列名
    :param kwargs:

        - fit_intercept: bool, 是否拟合截距项，默认为 False

    :return: pd.DataFrame, 新增 returns 列
    """
    from sklearn.linear_model import LinearRegression

    df = df.copy()
    fit_intercept = kwargs.get("fit_intercept", False)

    ret = []
    for dt, dfg in df.groupby("dt"):
        dfg = dfg.copy().dropna(subset=[factor, target])
        if dfg.empty or len(dfg) < 5:
            ret.append([dt, 0])
            logger.warning(f"{dt} has no enough data, only {len(dfg)} rows")
            continue

        x = dfg[factor].values.reshape(-1, 1)
        y = dfg[target].values.reshape(-1, 1)
        model = LinearRegression(fit_intercept=fit_intercept).fit(x, y)
        ret.append([dt, model.coef_[0][0]])

    dft = pd.DataFrame(ret, columns=["dt", "returns"])
    return dft


def feature_sectional_corr(df, factor, target="n1b", method="pearson", **kwargs):
    """计算因子特征截面相关性（IC）

    :param df：数据，DateFrame格式
    :param factor：因子列名，一般采用F#开头的列
    :param target：目标列名，一般为n1b
    :param method：{'pearson', 'kendall', 'spearman'} or callable

            * pearson : standard correlation coefficient
            * kendall : Kendall Tau correlation coefficient
            * spearman : Spearman rank correlation
            * callable: callable with input two 1d ndarrays and returning a float

    :return：df，res: 前者是每日相关系数结果，后者是每日相关系数的统计结果
    """
    from czsc.utils import single_linear

    df = df.copy()
    corr = []
    for dt, dfg in df.groupby("dt"):
        dfg = dfg.copy().dropna(subset=[factor, target])

        if dfg.empty or len(dfg) < 5:
            corr.append([dt, 0])
            logger.warning(f"{dt} has no enough data, only {len(dfg)} rows")
        else:
            c = dfg[factor].corr(dfg[target], method=method)
            corr.append([dt, c])

    dft = pd.DataFrame(corr, columns=["dt", "corr"])

    res = {
        "factor": factor,
        "target": target,
        "method": method,
        "IC均值": 0,
        "IC标准差": 0,
        "ICIR": 0,
        "IC胜率": 0,
        "累计IC回归R2": 0,
        "累计IC回归斜率": 0,
    }
    if dft.empty:
        return dft, res

    dft = dft[~dft["ic"].isnull()].copy()
    ic_avg = dft["ic"].mean()
    ic_std = dft["ic"].std()

    res["IC均值"] = round(ic_avg, 4)
    res["IC标准差"] = round(ic_std, 4)
    res["ICIR"] = round(ic_avg / ic_std, 4) if ic_std != 0 else 0
    if ic_avg > 0:
        res["IC胜率"] = round(len(dft[dft["ic"] > 0]) / len(dft), 4)
    else:
        res["IC胜率"] = round(len(dft[dft["ic"] < 0]) / len(dft), 4)

    lr_ = single_linear(y=dft["ic"].cumsum().to_list())
    res.update({"累计IC回归R2": lr_["r2"], "累计IC回归斜率": lr_["slope"]})
    return dft, res



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/6/25 18:52
"""
import time
import json
import requests
import pandas as pd
import tushare as ts
from deprecated import deprecated
from datetime import datetime
from typing import List
from functools import partial
from loguru import logger
from tenacity import retry, stop_after_attempt, wait_random
from czsc.objects import RawBar, Freq


# 数据频度 ：支持分钟(min)/日(D)/周(W)/月(M)K线，其中1min表示1分钟（类推1/5/15/30/60分钟）。
# 对于分钟数据有600积分用户可以试用（请求2次），正式权限请在QQ群私信群主或积分管理员。
freq_map = {
    Freq.F1: "1min",
    Freq.F5: "5min",
    Freq.F15: "15min",
    Freq.F30: "30min",
    Freq.F60: "60min",
    Freq.D: "D",
    Freq.W: "W",
    Freq.M: "M",
}
freq_cn_map = {
    "1分钟": Freq.F1,
    "5分钟": Freq.F5,
    "15分钟": Freq.F15,
    "30分钟": Freq.F30,
    "60分钟": Freq.F60,
    "日线": Freq.D,
}
dt_fmt = "%Y-%m-%d %H:%M:%S"
date_fmt = "%Y%m%d"


@deprecated(version="1.0.0", reason="不再推荐使用，后续随整个 data 模块一起删除")
class TushareProApi:
    __token = ""
    __http_url = "http://api.waditu.com"

    def __init__(self, token, timeout=30):
        """
        Parameters
        ----------
        token: str
            API接口TOKEN，用于用户认证
        """
        self.__token = token
        self.__timeout = timeout

    @retry(stop=stop_after_attempt(10), wait=wait_random(1, 5))
    def query(self, api_name, fields="", **kwargs):
        if api_name in ["__getstate__", "__setstate__"]:
            return pd.DataFrame()

        req_params = {"api_name": api_name, "token": self.__token, "params": kwargs, "fields": fields}

        res = requests.post(self.__http_url, json=req_params, timeout=self.__timeout)
        if res:
            result = json.loads(res.text)
            if result["code"] != 0:
                logger.warning(f"{req_params}: {result}")
                raise Exception(result["msg"])

            data = result["data"]
            columns = data["fields"]
            items = data["items"]
            return pd.DataFrame(items, columns=columns)
        else:
            return pd.DataFrame()

    def __getattr__(self, name):
        return partial(self.query, name)


try:
    from tushare.util import upass

    pro = TushareProApi(upass.get_token(), timeout=60)
except:
    print("Tushare Pro 初始化失败")


@deprecated(reason="统一到 ts_connector 中", version="1.0.0")
def format_kline(kline: pd.DataFrame, freq: Freq) -> List[RawBar]:
    """Tushare K线数据转换

    :param kline: Tushare 数据接口返回的K线数据
    :param freq: K线周期
    :return: 转换好的K线数据
    """
    bars = []
    dt_key = "trade_time" if "分钟" in freq.value else "trade_date"
    kline = kline.sort_values(dt_key, ascending=True, ignore_index=True)
    records = kline.to_dict("records")

    for i, record in enumerate(records):
        if freq == Freq.D:
            vol = int(record["vol"] * 100) if record["vol"] > 0 else 0
            amount = int(record.get("amount", 0) * 1000)
        else:
            vol = int(record["vol"]) if record["vol"] > 0 else 0
            amount = int(record.get("amount", 0))

        # 将每一根K线转换成 RawBar 对象
        bar = RawBar(
            symbol=record["ts_code"],
            dt=pd.to_datetime(record[dt_key]),
            id=i,
            freq=freq,
            open=record["open"],
            close=record["close"],
            high=record["high"],
            low=record["low"],
            vol=vol,  # 成交量，单位：股
            amount=amount,  # 成交额，单位：元
        )
        bars.append(bar)
    return bars


@deprecated(version="1.0.0", reason="不再推荐使用，后续随整个 data 模块一起删除")
def get_kline(
    ts_code: str,
    start_date: [datetime, str],
    end_date: [datetime, str],
    asset: str = "E",
    freq: Freq = Freq.F1,
    fq: str = "qfq",
) -> List[RawBar]:
    """
    通用行情接口: https://tushare.pro/document/2?doc_id=109

    :param ts_code:
    :param asset:
    :param freq:
    :param start_date:
    :param end_date:
    :param fq:
    :return:
    """
    start_date = pd.to_datetime(start_date)
    end_date = pd.to_datetime(end_date)
    if "分钟" in freq.value:
        start_date = start_date.strftime(dt_fmt)
        end_date = end_date.strftime(dt_fmt)
    else:
        start_date = start_date.strftime(date_fmt)
        end_date = end_date.strftime(date_fmt)

    df = ts.pro_bar(ts_code=ts_code, adj=fq, asset=asset, freq=freq_map[freq], start_date=start_date, end_date=end_date)
    bars = format_kline(df, freq)
    if bars and bars[-1].dt < pd.to_datetime(end_date) and len(bars) == 8000:
        print(f"获取K线数量达到8000根，数据获取到 {bars[-1].dt}，目标 end_date 为 {end_date}")
    return bars



# tests/test_cross_sectional_strategy.py
import pytest
import pandas as pd
from czsc.eda import cross_sectional_strategy


@pytest.fixture
def sample_data():
    data = {
        "dt": [
            "2023-01-01",
            "2023-01-02",
            "2023-01-03",
            "2023-01-04",
            "2023-01-05",
            "2023-01-06",
            "2023-01-07",
            "2023-01-08",
            "2023-01-09",
            "2023-01-10",
        ]
        * 5,
        "symbol": ["A"] * 10 + ["B"] * 10 + ["C"] * 10 + ["D"] * 10 + ["E"] * 10,
        "factor": list(range(1, 51)),
    }
    return pd.DataFrame(data)


def test_cross_sectional_strategy_positive(sample_data):
    result = cross_sectional_strategy(sample_data, factor="factor", long=0.5, short=0.5, factor_direction="positive")
    assert "weight" in result.columns
    assert result["weight"].sum() == 0  # Long and short positions should balance out


def test_cross_sectional_strategy_negative(sample_data):
    result = cross_sectional_strategy(sample_data, factor="factor", long=0.5, short=0.5, factor_direction="negative")
    assert "weight" in result.columns
    assert result["weight"].sum() == 0  # Long and short positions should balance out
    print(result)


def test_cross_sectional_strategy_negative_norm(sample_data):
    result = cross_sectional_strategy(
        sample_data, factor="factor", long=0.5, short=0.5, factor_direction="negative", norm=False
    )
    assert "weight" in result.columns
    assert result["weight"].sum() == 0  # Long and short positions should balance out
    print(result)


def test_cross_sectional_strategy_no_positions(sample_data):
    result = cross_sectional_strategy(sample_data, factor="factor", long=0, short=0)
    assert "weight" in result.columns
    assert result["weight"].sum() == 0  # No positions should be taken


def test_cross_sectional_strategy_invalid_factor(sample_data):
    with pytest.raises(AssertionError):
        cross_sectional_strategy(sample_data, factor="invalid_factor", long=0.5, short=0.5)


def test_cross_sectional_strategy_invalid_factor_direction(sample_data):
    with pytest.raises(AssertionError):
        cross_sectional_strategy(sample_data, factor="factor", long=0.5, short=0.5, factor_direction="invalid")


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/3/5 20:45
describe: CZSC投研数据共享接口

投研数据共享说明（含下载地址）：https://s0cqcxuy3p.feishu.cn/wiki/wikcnzuPawXtBB7Cj7mqlYZxpDh
"""
import os
import czsc
import glob
import pandas as pd
from datetime import datetime


# 投研共享数据的本地缓存路径，需要根据实际情况修改
cache_path = os.environ.get("czsc_research_cache", r"D:\CZSC投研数据")
if not os.path.exists(cache_path):
    raise ValueError(
        f"请设置环境变量 czsc_research_cache 为投研共享数据的本地缓存路径，当前路径不存在：{cache_path}。\n\n"
        f"投研数据共享说明（含下载地址）：https://s0cqcxuy3p.feishu.cn/wiki/wikcnzuPawXtBB7Cj7mqlYZxpDh"
    )


def get_groups():
    """获取投研共享数据的分组信息

    :return: 分组信息
    """
    return ["A股主要指数", "A股场内基金", "中证500成分股", "期货主力"]


def get_symbols(name, **kwargs):
    """获取指定分组下的所有标的代码

    :param name: 分组名称，可选值：'A股主要指数', 'A股场内基金', '中证500成分股', '期货主力'
    :param kwargs:
    :return:
    """
    if name.upper() == "ALL":
        files = glob.glob(os.path.join(cache_path, "*", "*.parquet"))
    else:
        files = glob.glob(os.path.join(cache_path, name, "*.parquet"))
    return [os.path.basename(x).replace(".parquet", "") for x in files]


def get_raw_bars(symbol, freq, sdt, edt, fq="前复权", **kwargs):
    """获取 CZSC 库定义的标准 RawBar 对象列表

    :param symbol: 标的代码
    :param freq: 周期，支持 Freq 对象，或者字符串，如
            '1分钟', '5分钟', '15分钟', '30分钟', '60分钟', '日线', '周线', '月线', '季线', '年线'
    :param sdt: 开始时间
    :param edt: 结束时间
    :param fq: 除权类型，投研共享数据默认都是后复权，不需要再处理
    :param kwargs:
    :return:
    """
    raw_bars = kwargs.get("raw_bars", True)
    kwargs["fq"] = fq
    file = glob.glob(os.path.join(cache_path, "*", f"{symbol}.parquet"))[0]
    freq = czsc.Freq(freq)
    kline = pd.read_parquet(file)
    if "dt" not in kline.columns:
        kline["dt"] = pd.to_datetime(kline["datetime"])
    kline = kline[(kline["dt"] >= pd.to_datetime(sdt)) & (kline["dt"] <= pd.to_datetime(edt))]
    if kline.empty:
        return []

    df = kline.copy()
    if symbol in ["SFIC9001", "SFIF9001", "SFIH9001"]:
        # 股指：仅保留 09:31 - 11:30, 13:01 - 15:00
        # 历史遗留问题，股指有一段时间，收盘时间是 15:15
        dt1 = datetime.strptime("09:31:00", "%H:%M:%S")
        dt2 = datetime.strptime("11:30:00", "%H:%M:%S")
        c1 = (df["dt"].dt.time >= dt1.time()) & (df["dt"].dt.time <= dt2.time())

        dt3 = datetime.strptime("13:01:00", "%H:%M:%S")
        dt4 = datetime.strptime("15:00:00", "%H:%M:%S")
        c2 = (df["dt"].dt.time >= dt3.time()) & (df["dt"].dt.time <= dt4.time())

        df = df[c1 | c2].copy().reset_index(drop=True)

    _bars = czsc.resample_bars(df, freq, raw_bars=raw_bars, base_freq="1分钟")
    return _bars



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/2/24 16:39
describe: data 模块的单元测试
"""
from czsc import data


def test_symbol_converter():
    jq_symbol = "300033.XSHE"
    gm_symbol = "SZSE.300033"
    ts_symbol = "300033.SZ"
    tdx_symbol = "0300033"

    assert data.gm_symbol_to_ts(gm_symbol) == ts_symbol
    assert data.gm_symbol_to_jq(gm_symbol) == jq_symbol
    assert data.gm_symbol_to_tdx(gm_symbol) == tdx_symbol

    assert data.jq_symbol_to_gm(jq_symbol) == gm_symbol
    assert data.jq_symbol_to_ts(jq_symbol) == ts_symbol
    assert data.jq_symbol_to_tdx(jq_symbol) == tdx_symbol

    assert data.ts_symbol_to_tdx(ts_symbol) == tdx_symbol
    assert data.ts_symbol_to_gm(ts_symbol) == gm_symbol
    assert data.ts_symbol_to_jq(ts_symbol) == jq_symbol

    assert data.ts_symbol_to_jq(ts_symbol) == jq_symbol
    assert data.ts_symbol_to_gm(ts_symbol) == gm_symbol
    assert data.ts_symbol_to_tdx(ts_symbol) == tdx_symbol

    assert data.tdx_symbol_to_ts(tdx_symbol) == ts_symbol
    assert data.tdx_symbol_to_gm(tdx_symbol) == gm_symbol
    assert data.tdx_symbol_to_jq(tdx_symbol) == jq_symbol
